//  bbsmount.cc for bbsmount - an tool for simple mounting in X11
//
//  Copyright (c) 2001 by Miroslav Jezbera, jezberam@phoenix.inf.upol.cz
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// (See the included file COPYING / GPL-2.0)
//

#ifdef HAVE_CONFIG_H
#   include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#if STDC_HEADERS
#   include <stdlib.h>
#   include <string.h>
#endif /* STDC_HEADERS */

#if HAVE_SIGNAL_H
#   include <signal.h>
#endif /* HAVE_SIGNAL_H */

#ifdef TIME_WITH_SYS_TIME
#   include <time.h>
#   include <sys/time.h>
#else /* !TIME_WITH_SYS_TIME */
#   include <time.h>
#endif /* !TIME_WITH_SYS_TIME */

#include <pthread.h>

#include "bbsmount.hh"
#include "version.h"

#include "images/notfound.xpm"

// Global variables
ToolWindow *AppWindow;
bool finish_checking, reconfigure;
int debug_level;

// Functions
void *
run_command(void *data)
{
    command_data_type *cd = (command_data_type *)data;
    int retval;

    cd->mp->Lock();
    retval = system(cd->command);
    if (retval == -1)
	perror("system");

    free(cd->command);
    cd->mp->Unlock();
    free(cd);
    pthread_exit(NULL);
    return NULL; // Never reached
}

int
run_command_in_foreground(const char *command, MountPoint &mp)
{
    int retval;

    if (strlen(command) == 0)
	return 0;
    retval = system(command);
    if (retval == -1)
	perror("system");

    return retval;
}

int
run_command_in_background(const char *command, MountPoint &mp)
{
    if (strlen(command) == 0)
	return 0;

    char *str = strdup(command);    // prevents change and compiler warnings
    command_data_type *data;
    int retval = 0;
    pthread_t run;
    pthread_attr_t attribs;

    data = (command_data_type *)malloc(sizeof(command_data_type));
    if (data == NULL) {
	fprintf(stderr, "bbsmount: Memory allocation failed!\n");
	return -1;
    }
    data->command = str;
    data->mp = &mp;
    if (pthread_attr_init(&attribs) != 0) {
	perror("pthread_attr_init");
	return -1;
    }
    if (pthread_attr_setdetachstate(&attribs, PTHREAD_CREATE_DETACHED) != 0) {
	perror("pthread_attr_setdetachstate");
	pthread_attr_destroy(&attribs);
	return -1;
    }
    if (pthread_create(&run, &attribs, run_command, (void *)data) != 0) {
	perror("pthread_create");
	retval = -1;
    }
    if (pthread_attr_destroy(&attribs) != 0)
	perror("pthread_attr_destroy");

    return retval;
}

set<int>
what_is_mounted(const map<string, int> &mount_points)
{
    set<int> result;
    map<string, int>::const_iterator pointer;
    char buffer[max_mtab_line_length], md[max_mount_device_point_size], mp[max_mount_device_point_size];
    FILE *mtab;

#if READ_MOUNT
    mtab = popen(MOUNT_COMMAND, "r");
    if (!mtab) {
	fprintf(stderr, "Can't read output from mount command!");
	return result;
    }
#else /* !READ_MOUNT */
    mtab = fopen(MTAB_FILE, "rt");
    if (!mtab) {
	fprintf(stderr, "Can't open mtab file!");
	return result;
    }
#endif /* !READ_MOUNT */
    
    while (!feof(mtab)) {
	if (fgets(buffer, max_mtab_line_length, mtab) > 0) {
#if READ_MOUNT
	    sscanf(buffer, "%s on %s", md, mp);
#else /* !READ_MOUNT */
	    sscanf(buffer, "%s %s", md, mp);
#endif /* !READ_MOUNT */
	    pointer = mount_points.find(string(mp));
	    if (pointer != mount_points.end())
		result.insert(pointer->second);
	}
    }
    
#if READ_MOUNT
    pclose(mtab);
#else /* !READ_MOUNT */
    fclose(mtab);
#endif /* !READ_MOUNT */
    return result;
}

bool
update_mount_points(const map<string, int> &mount_points)
{
    set<int> result;
    size_t count = AppWindow->GetMountPointsCount();
    bool change = false;
    int counter;

    result = what_is_mounted(mount_points);
    for (counter = 0; counter < (int)count; counter++)
	if (AppWindow->GetMountPoint(counter).IsMounted() != (result.find(counter) != result.end())) {
	    AppWindow->GetMountPoint(counter).SetMounted(result.find(counter) != result.end());
	    change = true;
	}
    return change;
}

void
create_notify_event(void)
{
    XEvent event;

    memset((void *)&event, 0, sizeof(XExposeEvent));
    event.type = Expose;		/* How can I notify my thread waiting in XNextEvent, that there is some work ? */
    event.xexpose.window = AppWindow->GetRootWindow();
    event.xexpose.display = AppWindow->getDisplay();
    event.xexpose.send_event = True;
    XSendEvent(AppWindow->getDisplay(), AppWindow->GetRootWindow(), False, NoEventMask, &event);
    XFlush(AppWindow->getDisplay());
}

void *
check_mounts(void *param)
{
#if !READ_MOUNT
    struct stat info;
    int mtab_size = 0;
    time_t mtab_time = time(NULL);
#endif /* !READ_MOUNT */
    int return_value, counter;
    map<string, int> mount_points;

    for (counter = 0; counter < (int)AppWindow->GetMountPointsCount(); counter++)
	mount_points[AppWindow->GetMountPoint(counter).GetMountPoint()] = counter;
    while (!finish_checking) {
#if READ_MOUNT
	if (update_mount_points(mount_points))
		create_notify_event();
#else /* !READ_MOUNT */
	stat(MTAB_FILE, &info);
	if ((difftime(info.st_mtime, mtab_time) > 0) || (mtab_size != info.st_size)) {
	    mtab_size = info.st_size;
	    mtab_time = info.st_mtime;
	    if (update_mount_points(mount_points))
		create_notify_event();
	}
#endif /* !READ_MOUNT */
	return_value = sleep(AppWindow->GetResources()->app.refresh_time);
	if (return_value != 0)
	    finish_checking = true;
    }

    pthread_exit(NULL);
    return NULL;
}

RETSIGTYPE
shutdown(int signal)
{
#if DEBUG
    if (debug_level >= dbg_all_work)
	printf("Stopping bbsmount.\n");
#endif
    finish_checking = true;
    reconfigure = false;
    AppWindow->Stop();
    create_notify_event();
}

ToolWindow::ToolWindow(int argc,char **argv,struct CMDOPTIONS *options) : Basewindow(argc,argv,options)
{
    XrmInitialize();
    resource = new Resource(this);
    CreateWindow(false);
    Configure(true);
    stop = false;
}

ToolWindow::~ToolWindow(void)
{
    XUnmapSubwindows(getDisplay(), rootwin);
    FreeAll();
    XUnmapWindow(getDisplay(), rootwin);

    /* destroy pixmaps */
    if (pixmap.frame)
	getImageControl()->removeImage(pixmap.frame);
    if (pixmap.draw)
	getImageControl()->removeImage(pixmap.draw);
    if (pixmap.draw_pressed)
	getImageControl()->removeImage(pixmap.draw_pressed);

    delete resource;
    XDestroyWindow(getDisplay(), rootwin);
    XFlush(getDisplay());
}
 
const MyImage &
ToolWindow::GetImage(const int imageidx) const
{
    return *images[imageidx];
}

const MountPoint &
ToolWindow::GetMountPoint(const int mntidx) const
{
    return mount_points[mntidx];
}

const PIXMAP &
ToolWindow::GetPixmaps(void) const
{
    return pixmap;
}

size_t
ToolWindow::GetMountPointsCount(void) const
{
    return mount_points.size();
}

MountPoint &
ToolWindow::GetMountPoint(const int mntidx)
{
    return mount_points[mntidx];
}

void
ToolWindow::FreeAll(void)
{
    vector<MyImage *>::iterator pointer = images.begin();
    vector<MountWindow *>::iterator mwpointer = mount_windows.begin();
    
    mount_points.clear();
    commands.clear();
    for (; pointer != images.end(); pointer++)
	delete *pointer;
    images.clear();
    for (; mwpointer != mount_windows.end(); mwpointer++)
	delete *mwpointer;
    mount_windows.clear();
}

void
ToolWindow::Configure(const bool first_time)
{
    vector<string>::const_iterator pointer, end;
    XpmImage new_image;
    int image_counter;

    if (!first_time) {
	FreeAll();

	resource->MountPoints();
    }
    commands = resource->GetCommands();
    mount_points = resource->GetMountPoints();
    pointer = resource->GetImages().begin();
    end = resource->GetImages().end();
    
    XpmCreateXpmImageFromData(notfound_xpm, &new_image, NULL);
    SetImage(0, new MyImage(getDisplay(), rootwin, &new_image));
    XpmFreeXpmImage(&new_image);
    image_counter = 1;

    for (; pointer != end; pointer++) {
	File imagefile(pointer->c_str(), resource->GetImagePrefix().c_str());
	MyImage *my_new_image;

	if (imagefile.Exist()) {
#if DEBUG
	    if (debug_level >= dbg_all_work)
		printf("Loading Xpm image %s.\n", imagefile.GetPath());
#endif /* DEBUG */
	    if (XpmReadFileToXpmImage(imagefile.GetPath(), &new_image, NULL) != XpmSuccess) {
		XpmCreateXpmImageFromData(notfound_xpm, &new_image, NULL);
#if DEBUG
		printf("Can't read image data from file %s!\n", imagefile.GetPath());
#endif /* DEBUG */
	    }
	}
	else {
	    XpmCreateXpmImageFromData(notfound_xpm, &new_image, NULL);
#if DEBUG
	    printf("Can't find image file %s!\n", pointer->c_str());
#endif /* DEBUG */
	}
	
	my_new_image = new MyImage(getDisplay(), rootwin, &new_image);
	SetImage(image_counter++, my_new_image);
	XpmFreeXpmImage(&new_image);
    }
    resource->FreeAll();
}

/*Display *
ToolWindow::GetDisplay(void)
{
    return dpy;
}*/

Resource *
ToolWindow::GetResources(void)
{
    return resource;
}

Window
ToolWindow::GetRootWindow(void)
{
    return rootwin;
}

void
ToolWindow::CreateWindow(const bool reconfigure)
{
    XSetWindowAttributes attrib;
    XWMHints wmhints;
    unsigned long create_mask;

    create_mask = CWBackPixmap | CWOverrideRedirect | CWCursor | CWEventMask; 

    drawarea.width = resource->app.column_width * resource->app.columns;
    drawarea.height = resource->app.row_height * resource->app.rows;

    rootgeometry.width = resource->app.column_width * resource->app.columns +
	resource->frame.bevelWidth * 2 + (resource->app.columns - 1) * resource->app.button_padding;
    rootgeometry.height = resource->app.row_height * resource->app.rows +
	resource->frame.bevelWidth * 2 + (resource->app.rows - 1) * resource->app.button_padding;

    if (resource->position.mask & XNegative) {
	resource->position.x = getCurrentScreenInfo()->getWidth() + resource->position.x - rootgeometry.width;
    }
    if (resource->position.mask & YNegative) {
	resource->position.y = getCurrentScreenInfo()->getHeight() + resource->position.y - rootgeometry.height;
    }
    if (withdrawn) {
	attrib.override_redirect = False;
	wmhints.initial_state = WithdrawnState;
    }
    else {
	attrib.override_redirect = True;
	wmhints.initial_state = NormalState;
    }

    attrib.background_pixmap = ParentRelative;
    attrib.cursor = getSessionCursor();
    attrib.event_mask = ButtonPressMask | ButtonReleaseMask | ExposureMask |
			FocusChangeMask | KeyPressMask | StructureNotifyMask;

    pixmap.frame = getImageControl()->renderImage(rootgeometry.width, rootgeometry.height, &resource->frame.texture);
    pixmap.draw = getImageControl()->renderImage(resource->app.column_width, resource->app.row_height, &resource->app.mount_texture);
    pixmap.draw_pressed = getImageControl()->renderImage(resource->app.column_width, resource->app.row_height, &resource->app.mount_texture_pressed);

    if (!reconfigure) {
	rootwin = XCreateWindow(getDisplay(), getCurrentScreenInfo()->getRootWindow(), resource->position.x, resource->position.y, rootgeometry.width,
		rootgeometry.height, 0, getCurrentScreenInfo()->getDepth(), InputOutput, getCurrentScreenInfo()->getVisual(), create_mask, &attrib);
    }
    else
	if (!withdrawn) {
	    XMoveResizeWindow(getDisplay(), rootwin, resource->position.x, resource->position.y,
		   rootgeometry.width, rootgeometry.height);
	}
	else {
	    XResizeWindow(getDisplay(), rootwin, rootgeometry.width, rootgeometry.height);
	}

    wmhints.flags = IconWindowHint | StateHint;
    wmhints.icon_window = rootwin;
    XSetStandardProperties(getDisplay(), rootwin, BBTOOL, BBTOOL, None,
	    getArgv(), getArgc(), NULL);
    XSetWMProtocols (getDisplay(), rootwin, &wm_delete_window, 1);
    XSetWMHints(getDisplay(), rootwin, &wmhints);

    if (!shape)
	XSetWindowBackgroundPixmap(getDisplay(), rootwin, pixmap.frame);

    if (!reconfigure) {
	XClearWindow(getDisplay(), rootwin);
	XMapWindow(getDisplay(), rootwin);  
    }
}

void
ToolWindow::CreateSubWindows(const bool reconfigure)
{
    int col = 0, row = 0;
    vector<MountPoint>::iterator pointer;
    vector<MountWindow *>::iterator mwpointer = mount_windows.begin();

    if (!reconfigure) {
	for (pointer = mount_points.begin(); pointer != mount_points.end(); pointer++) {
	    if (col == resource->app.columns) {
		if (++row == resource->app.rows)
		    break;
		col = 0;
	    }
	    mount_windows.push_back(new MountWindow(&*pointer, getDisplay(), rootwin, resource->frame.bevelWidth +
		    col * resource->app.column_width + resource->app.button_padding * col,
		    resource->frame.bevelWidth + row * resource->app.row_height +
		    resource->app.button_padding * row, resource->app.column_width,
		    resource->app.row_height));
	    mount_windows.back()->Draw();
	    col++;
	}
    }
    else
	if (!withdrawn) {
	    for (; mwpointer != mount_windows.end(); mwpointer++)
		delete *mwpointer;
	    mount_windows.clear();
	    for (pointer = mount_points.begin(); pointer != mount_points.end(); pointer++) {
		if (col == resource->app.columns) {
		    if (++row == resource->app.rows)
			break;
		    col = 0;
		}
		mount_windows.push_back(new MountWindow(&*pointer, getDisplay(), rootwin, resource->frame.bevelWidth +
			col * resource->app.column_width + resource->app.button_padding * col,
			resource->frame.bevelWidth + row * resource->app.row_height +
			resource->app.button_padding * row, resource->app.column_width,
			resource->app.row_height));
		col++;
	    }
	}
	else {
	    for (; mwpointer != mount_windows.end(); mwpointer++)
		delete *mwpointer;
	    mount_windows.clear();
	    for (pointer = mount_points.begin(); pointer != mount_points.end(); pointer++) {
		if (col == resource->app.columns) {
		    if (++row == resource->app.rows)
			break;
		    col = 0;
		}
		mount_windows.push_back(new MountWindow(&*pointer, getDisplay(), rootwin, resource->frame.bevelWidth +
			col * resource->app.column_width + resource->app.button_padding * col,
			resource->frame.bevelWidth + row * resource->app.row_height +
			resource->app.button_padding * row, resource->app.column_width,
			resource->app.row_height));
		col++;
	    }
	}

    if (!reconfigure)
	XMapSubwindows(getDisplay(), rootwin);
}

void
ToolWindow::EventLoop(void)
{
    XEvent event;
    vector<MountWindow *>::iterator pointer;
    unsigned int command;

    while (!stop) {
	XNextEvent(getDisplay(), &event);
	switch (event.type) {
	    case Expose:
		if (stop)
		    return;
		if (event.xexpose.window == rootwin) {
		    if (event.xexpose.send_event == True) {
			for (pointer = mount_windows.begin(); pointer != mount_windows.end(); pointer++)
				(*pointer)->Update();
		    }
		    else
			Draw(event.xexpose.x, event.xexpose.y, event.xexpose.width, event.xexpose.height);
		}
		else
		    for (pointer = mount_windows.begin(); pointer != mount_windows.end(); pointer++)
			if (event.xexpose.window == (*pointer)->GetWindow()) {
				(*pointer)->Draw(event.xexpose.x, event.xexpose.y, event.xexpose.width,
					event.xexpose.height);
				break;
			}
		break;
	    case ButtonPress:
		for (pointer = mount_windows.begin(); pointer != mount_windows.end(); pointer++)
		    if (event.xbutton.window == (*pointer)->GetWindow()) {
			command = (*pointer)->GetMountPoint().GetCommand(event.xbutton.button, event.xbutton.state);
			if (command) {
			    (*pointer)->SetPressed(true);
			    char *comm = (char *)malloc(sizeof(char) * (commands[command].length() + (*pointer)->GetMountPoint().GetMountPoint().length()));
			    sprintf(comm, commands[command].c_str(), (*pointer)->GetMountPoint().GetMountPoint().c_str());
#if DEBUG
			    if (debug_level >= dbg_all_work)
				printf("Executing command: `%s`\n", comm);
#endif /* DEBUG */
#if RUN_IN_BACKGROUND
			    run_command_in_background(comm, (*pointer)->GetMountPoint());
#else /* !RUN_IN_BACKGROUND */
			    run_command_in_foreground(comm, (*pointer)->GetMountPoint());
#endif /* !RUN_IN_BACKGROUND */
			    free(comm);
			}
		    }
		break;
	    case ButtonRelease:
		for (pointer = mount_windows.begin(); pointer != mount_windows.end(); pointer++)
		    if (event.xbutton.window == (*pointer)->GetWindow())
			(*pointer)->SetPressed(false);
		break;
	    case ConfigureNotify:
#if DEBUG
		if (debug_level >= dbg_all_work)
		    printf("ConfigureNotify detected.\n");
#endif /* DEBUG */
		break;
	}
    }
}

void
ToolWindow::Draw(void)
{
    XClearWindow(getDisplay(), rootwin);
}

void
ToolWindow::Draw(const int x, const int y, const unsigned int width, const unsigned int height)
{
    XClearArea(getDisplay(), rootwin, x, y, width, height, False);
}

void
ToolWindow::SetCommand(const int cmdidx, const string command)
{
    if (cmdidx == (int)commands.size())
	commands.push_back(command);
    else
	commands[cmdidx] = command;
}

void
ToolWindow::SetImage(const int imgidx, MyImage *image)
{
    if (imgidx == (int)images.size())
	images.push_back(image);
    else
	images[imgidx] = image;
}

void
ToolWindow::SetMountPoint(const int mntidx, const MountPoint &mntpoint)
{
    if (mntidx == (int)mount_points.size())
	mount_points.push_back(mntpoint);
    else
	mount_points[mntidx] = mntpoint;
}

void
ToolWindow::Stop(void)
{
    stop = true;
}

int
main(int argc, char *argv[])
{
    int result;
    pthread_t check_thread;
    void *thread_return;
    struct CMDOPTIONS options;

    if (XInitThreads() == 0) {
    	perror("bbsmount: Can't initialize X11!");
    	exit(EXIT_FAILURE);
    }

    read_options(argc, argv, options);
    AppWindow = new ToolWindow(argc, argv, &options);

    AppWindow->CreateSubWindows(false);

#ifdef HAVE_SIGACTION
    struct sigaction act;

    act.sa_handler = shutdown;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;

    sigaction(SIGINT, &act, 0);
    sigaction(SIGTERM, &act, 0);
#else /* !HAVE_SIGACTION */
    signal(SIGINT, shutdown);
    signal(SIGTERM, shutdown);
#endif /* !HAVE_SIGACTION */

    do {
	finish_checking = false;
	reconfigure = false;

	result = pthread_create(&check_thread, NULL, check_mounts, NULL);
	if (result != 0) {
	    perror("bbsmount: Can't create thread!");
	    exit(EXIT_FAILURE);
	}

	AppWindow->EventLoop();

	finish_checking = true;
	result = pthread_join(check_thread, &thread_return);
	if (result != 0) {
	    perror("bbsmount: Can't join thread!");
	    exit(EXIT_FAILURE);
	}
    } while (reconfigure);
    
    delete AppWindow;

    return EXIT_SUCCESS;
}
